/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "project.h"

plugconstructor::plugconstructor(int name, attribute *attribs, list<symbolname> *symbols, plugtype *type) : symbolconstructor(name, attribs)
{
  plugconstructor::symbols = symbols;
  revlist(symbols);
  plugconstructor::type = type;
}

symbol *plugconstructor::_create(symbol *parent)
{
  return new plug(name, attribs, parent, this); // Not very orthodox cast!
}

plug::plug(int name, attribute *attribs, symbol *parent, plugconstructor *constr) : symbol(name, attribs, parent)
{
  int allpins = 1;
  int somenotlocal = 0;
  int numsymbols = 0;
  type = constr->type;
  symbols = (list<symbol> *) NULL;

  for (list<symbolname> *l = constr->symbols; l != NULL; l = l->next)
  {
    numsymbols++;
    if (l->elem != NULL)
    {
      symbol *s = parent->getlocalsymbol(l->elem);
      if (s == NULL) 
      {
	s = parent->getsymbol(l->elem);
	if (s == NULL) 
        {
	  constr->createerror(TYPE_ERROR, "Can't find symbol %s. Using dummy.", symbtable.tostring(l->elem));

          if (parent->ToComponent() != NULL)
            s = nullpin(attribs, parent->ToComponent());
          else
            s = nullsymbol(attribs, parent); // Insert a dummy symbol...
        }
        else
          somenotlocal = 1;
      }
      if (s->ToPin() == NULL) allpins = 0;
      symbols = new list<symbol>(s, symbols);
    }
    else
    {
      if (parent->ToComponent() != NULL)
        // We use nullpin because it has to act like a pin in a bus
        symbols = new list<symbol>(nullpin(attribs, parent->ToComponent()), symbols);
      else
        symbols = new list<symbol>(nullsymbol(attribs, parent), symbols); // Insert a dummy symbol...
    }
  }

  if (type != NULL && type->NumSymbols() != numsymbols)
    constr->createerror(TYPE_ERROR, "Tried to create %i pin plug %s of %i pin type %s", numsymbols, symbtable.tostring(name), type->NumSymbols(), symbtable.tostring(type->GetName()));

  revlist(symbols);

  if (allpins && somenotlocal) constr->createerror(TYPE_WARNING, "This pin plug has some references to non-children symbols. Simulation may not work correctly!");
}

plug::plug(int name, attribute *attribs, symbol *parent) : symbol(name, attribs, parent)
{
  symbols = (list<symbol> *) NULL;
  type = 0;
}

symbolwidget *plug::getwidget()
{
  return new plugwidget(getattributes(), this);
}

symbol *plug::_getlocalsymbol(symbolname *s)
{
  if (type != NULL)
  {
    list<symbolname> *sn = type->GetSymbolNames();
    for (list<symbol> *l = symbols; l != NULL && sn != NULL; l = l->next, sn = sn->next)
      if (s->isprefix(sn->elem))
      {
	symbolname *fnd = sn->elem;
	while (fnd != NULL) fnd = fnd->next, s = s->next;
        return l->elem->getlocalsymbol(s);
      }
  }
    
  int n = *(s->elem);
  if (n < 0) return (symbol *) NULL;
  for (list<symbol> *l = symbols; l != NULL; n--, l = l->next)
    if (n == 0) return l->elem->getlocalsymbol(s->next);

  return (symbol *) NULL;
}

plugtype::plugtype(int name, list<symbolname> *symbolnames)
{
  plugtype::name = name;
  plugtype::symbolnames = symbolnames;
  numsymbols = 0;
  for (list<symbolname> *l = symbolnames; l != NULL; l = l->next)
    numsymbols++;
}

busconstructor::busconstructor(int name, attribute *attribs, list<symbolname> *plugnames, int _mixed) : symbolconstructor(name, attribs)
{
  busconstructor::plugnames = plugnames;
  revlist(plugnames);
  mixed = _mixed;
}

symbol *busconstructor::_ccreate(macrocomponent *parent)
{
  return new bus(name, attribs, parent, this);
}

bus::bus(int name, attribute *attribs, macrocomponent *parent, busconstructor *constr) : plug(name, attribs, parent)
{
  plugs = (list<plug> *) NULL;

  for (list<symbolname> *l = constr->plugnames; l != NULL; l = l->next) 
  { // Get plugs from names.
    symbol *s = parent->getlocalsymbol(l->elem);
    if (s == NULL) 
    {
      constr->createerror(TYPE_ERROR, "Unknown plug %s while creating bus %s", symbtable.tostring(l->elem), symbtable.tostring(parent->GetType()->name));
      continue;
    }
    plug *p = s->ToPlug();
    if (p == NULL) 
    {
      constr->createerror(TYPE_ERROR, "Not a plug %s while creating bus %s", symbtable.tostring(l->elem), symbtable.tostring(parent->GetType()->name));
      continue;
    }
    plugs = new list<plug>(p, plugs);
  }

  type = plugs->elem->type;

  list<list<symbol> > *curpins = (list<list<symbol> > *) NULL;
  for (list<plug> *l = plugs; l != NULL; l = l->next)
  { // Prepare curpins for next step.
    curpins = new list<list<symbol> >(l->elem->symbols, curpins);
    if (!constr->mixed && type != l->elem->type) 
    {
      if (type != NULL && l->elem->type != NULL) constr->createerror(TYPE_ERROR, "%s of type %s on bus of type %s. You could make a mixedbus.", symbtable.tostring(l->elem->getname()), symbtable.tostring(l->elem->type->GetName()), symbtable.tostring(type->GetName()));
      else if (type != NULL) constr->createerror(TYPE_ERROR, "%s untyped on bus of type %s. You could make a mixedbus.", symbtable.tostring(l->elem->getname()), symbtable.tostring(type->GetName()));
      else constr->createerror(TYPE_ERROR, "%s of type %s on untyped bus. You could make a mixedbus.", symbtable.tostring(l->elem->getname()), symbtable.tostring(l->elem->type->GetName()));
    }
  }

  list<net> *nets;
  while (curpins->elem != NULL)
  { // Generate the necessary nets.
    list<pin> *netpins = (list<pin> *) NULL;
    for (list<list<symbol> > *l = curpins; l != NULL; l = l->next)
    {
      pin *p;
      if (l->elem == NULL) 
      {
	constr->createerror(TYPE_ERROR, "Tried to connect pins of different length in bus %s", symbtable.tostring(name));
	p = nullpin(attribs, parent);
      }
      else p = l->elem->elem->ToPin();
      if (p == NULL)
      { 
	constr->createerror(TYPE_ERROR, "Tried to make a bus %s with non pin element %s", symbtable.tostring(name), l->elem->elem->getfullnametext());
	p = nullpin(attribs, parent);
      }
      netpins = new list<pin>(p, netpins);
      delete l->elem;
      l->elem = l->elem->next;
    }
    nets = new list<net>(new net(symbtable.toid(""), attribs, netpins, this, parent), nets);
  }

  for (list<list<symbol> > *l = curpins; l != NULL; l = l->next)
  {
    if (l->elem != NULL) constr->createerror(TYPE_ERROR, "Tried to connect pins of different length in bus %s", symbtable.tostring(name));
  }

  symbols = (list<symbol> *) nets;
  revlist(symbols);
}
